;;;                 -*- Mode: Lisp; Package: BOOMS-APP -*-                  ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         iconedit.lsp                                              ;;;
;;; Description:  BOOMS iconic nodes editing                                ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Date:         29-12-96                                                  ;;;
;;; Package:      BOOMS-APP                                                 ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  I C O N E D I T  file                                                  ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :booms-app)


;;;-------------------------------------------------------------------------;;;
;;;  I C O N - E D I T  dialog variables                                    ;;;
;;;-------------------------------------------------------------------------;;;

;;; The dialog itself.
(defvar *ICON-EDIT-DIALOG*         nil)
;;; This will be set to t wheneved icon-edit is called, so no double call
;;; happens - during initializing of an icon.
(defvar *ICON-EDIT-CALLED*         nil)
;;; The node used to hold the arguments.
(defvar *ICON-EDIT-ICON*           nil)
;;; Is the dialog running?
(defvar *ICON-EDIT-DIALOG-RUNNING* nil)
;;; The dialog tab widget.
(defvar *ICON-EDIT-THE-TAB*        nil)
;;; The last tab-id used for building a tab page - an integer.
(defvar *LAST-TAB-ID*              nil)
;;; The tab ids and infos (each a list of both).
(defvar *TAB-IDS/INOFS*            nil)

;;; Initialize the dialog.
(defun INITIALIZE-ICON-DIALOG ()
  (setf *icon-edit-dialog* nil)
  (setf *last-tab-id*      nil)
  (setf *tab-ids/inofs*    nil)
  (make-icon-edit-dialog)
  (make-iconic-node-tab)
  (set-dialog-item-value *icon-edit-the-tab* nil))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N - E D I T  function definition                                 ;;;
;;;-------------------------------------------------------------------------;;;

;;; This will return the list of tab ids for some icon.
(defmethod ICON-TAB-IDS ((x t)) nil)

;;; Return the node page for iconic-node and the colors page for all icons.
(defmethod ICON-TAB-IDS ((x iconic-node))
  (list* :node (call-next-method)))

;;; Edit an icon.
(defmethod ICON-EDIT ((icon icon) &optional socket default-tab cursor-center)
  (setf *icon-edit-called* t)
  (let* ((pos               #.(make-position 0 0))
         (icon-doc          (icon-doc icon))
         (tab-ids           (icon-tab-ids icon))
         (dlg-width         (window-exterior-width  *icon-edit-dialog*))
         (dlg-height        (window-exterior-height *icon-edit-dialog*))
         (*prevent-hilight* t)
         (*icon-edit-icon*  icon))
    (when tab-ids
      ;; First of all - select & bring to front this icon if it exists (not
      ;; called from icon-initialize).
      (when (member icon (icon-list icon-doc))
        (select-one-icon icon-doc icon))
      (unhilight-current-icon icon-doc)
      (unless cursor-center
        ;; Make the dialog stick to the icon, but not outside the screen.
        (nwindow-interior-top-left *booms-app* pos)
        (nposition+ pos (window-interior-top-left (get-booms-pane)))
        (nposition+ pos (window-interior-top-left (window-parent icon-doc)))
        (nposition+ pos (window-interior-top-left icon-doc))
        (nposition- pos (scroll-position icon-doc))
        (nposition+ pos (nnbox-top-right (icon-box icon)))
        (when (> (+ (position-x pos) dlg-width) (get-screen-size-x))
          (decf (position-x pos) (+ dlg-width (icon-width icon))))
        (when (> (+ (position-y pos) dlg-height) (get-screen-size-y))
          (decf (position-y pos) (- dlg-height (icon-height icon))))
        (move-window *icon-edit-dialog* pos))
      ;; Set the active pages.
      (set-active-tabs tab-ids default-tab)
      ;; Setup page items and values.
      (dolist (tab-id tab-ids)
        (icon-edit-setup-page icon tab-id socket))
      ;; Pop up the dialog.
      (let* ((*icon-edit-dialog-running* t)
             (result (my-pop-up-dialog *icon-edit-dialog*
                                       (and cursor-center :cursor))))
        (dolist (tab-id tab-ids)
          (icon-edit-done-page icon tab-id result))
        ;; Redraw the screen for any possible changes.
        (do-redraws (icon-doc *icon-edit-icon*))
        ;; Return value - the result of the dialog
        result))))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N - E D I T  basic dialog constructing                           ;;;
;;;-------------------------------------------------------------------------;;;

;;; The initial creation of the basic dialog box - only the tab control and the
;;; OK/Cancel buttons.
(defun MAKE-ICON-EDIT-DIALOG ()
  (unless *icon-edit-dialog*
    (setf *icon-edit-dialog*
          (open-dialog
           (list
            (make-dialog-item
              :widget        'default-button
              :title         "~OK"
              :box           (make-box 7 155 117 180)
              :3d-border     t
              :tabstop       t
              :groupstart    t
              :set-value-fn  'return-t-from-pop-up-dialog)
            (make-dialog-item
              :widget        'cancel-button
              :title         "Cancel"
              :box           (make-box 123 155 233 180)
              :3d-border     t
              :tabstop       t
              :groupstart    nil
              :set-value-fn  'return-nil-from-pop-up-dialog)
            (setf *icon-edit-the-tab*
             (make-dialog-item
              :widget        'tab-control
              :name          :the-tab
              :value         nil
              :single-line-p t
              :fixed-width-p nil
              :box           (make-box 0 0 240 150)
              :tabstop       t
              :groupstart    t
              :tab-height    20
              :range         nil)))
           'dialog *booms-app*
           :title             "Edit Icon"
           :window-state      nil
           :window-border     :dialog-box
           :left-attachment   nil
           :top-attachment    nil
           :right-attachment  nil
           :bottom-attachment nil
           :user-movable      t
           :user-resizable    nil
           :user-closable     nil
           :user-shrinkable   nil
           :user-scrollable   nil
           :background-color  light-gray
           :pop-up-p          t
           :window-interior   (make-box 0 0 240 185)))))


;;;-------------------------------------------------------------------------;;;
;;;  T A B - C O N T R O L  handling                                        ;;;
;;;-------------------------------------------------------------------------;;;

;;; Set some pages active, hide the rest.  default is a default page to
;;; activate.
(defun SET-ACTIVE-TABS (ids &optional default-tab)
  (let ((old-val (dialog-item-value *icon-edit-the-tab*)))
    (dolist (id/info *tab-ids/inofs*)
      (tab-control-remove-tab *icon-edit-the-tab* (first id/info)))
    (dolist (id/info *tab-ids/inofs*)
      (when (member (first id/info) ids)
        (tab-control-restore-tab
         *icon-edit-the-tab* (second id/info) :index -1)))
    (set-dialog-item-value *icon-edit-the-tab* nil)
    (set-dialog-item-value
     *icon-edit-the-tab*
     (or default-tab
         (if (member old-val ids)
           old-val
           (pc::tab-info-id
            (first (dialog-item-range *icon-edit-the-tab*))))))))

;;; Add a tab to the tab-colntrol widget.
(defun ADD-ICON-EDIT-TAB (tab-id)
  (unless (member tab-id *tab-ids/inofs* :key #'first)
    (tab-control-add-tab *icon-edit-the-tab*
                         tab-id
                         (capitalize-object tab-id))
    (pushlast (list tab-id (tab-control-tab *icon-edit-the-tab* tab-id))
              *tab-ids/inofs*))
  (setf *last-tab-id* tab-id))

;;; Add a widget to the last added tab page.
(defun ADD-ICON-EDIT-TAB-WIDGET (&rest widget-specs)
  (let ((widget (apply #'make-dialog-item
                       :tab-control (list :the-tab *last-tab-id*)
                       widget-specs)))
    (add-widget widget *icon-edit-dialog* t t)
    (tab-control-add-widget *icon-edit-the-tab* *last-tab-id* widget)
    widget))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N I C - N O D E  tab definitions                                 ;;;
;;;-------------------------------------------------------------------------;;;

;;; List of saved values - conses of nodes & constant.
(defvar *NODE-EDIT-VALUES*          nil)

;;; The original node arguments.
(defvar *NODE-SAVED-ARGUMENTS*      nil)

;;; The node page widgets.
(defvar *NODE-EDIT-ARGUMENT-LIST*   nil)
(defvar *NODE-EDIT-ADD-BUTTON*      nil)
(defvar *NODE-EDIT-DEL-BUTTON*      nil)
(defvar *NODE-EDIT-NO-VALUE-BUTTON* nil)
(defvar *NODE-EDIT-NODE-BUTTON*     nil)
(defvar *NODE-EDIT-NODE-LIST*       nil)
(defvar *NODE-EDIT-CONSTANT-BUTTON* nil)
(defvar *NODE-EDIT-CONSTANT-TEXT*   nil)

;;; Make the :node tab page.
(defun MAKE-ICONIC-NODE-TAB ()
  (add-icon-edit-tab :node)
  ;; Add the widgets in reverse order for correct tab order of the widgets.
  (progrev
   (add-icon-edit-tab-widget
     :widget       'static-text
     :value        "~Argument"
     :box          (make-box 10 25 90 45)
     :tabstop      nil
     :groupstart   t)
   (setf *node-edit-argument-list*
    (add-icon-edit-tab-widget
     :widget       'single-item-list
     :value        nil
     :box          (make-box 10 45 60 123)
     :3d-border    t
     :tabstop      t
     :groupstart   nil
     :range        nil
     :key          #'(lambda (x) (format nil "~2@a" (1+ x)))
     :font         (make-font :modern :courier\ new 24 '(:bold))
     :set-value-fn 'icon-edit-argument-list-set-fn))
   (setf *node-edit-add-button*
    (add-icon-edit-tab-widget
     :widget       'button
     :title        "+"
     :box          (make-box 10 125 35 145)
     :3d-border    t
     :tabstop      t
     :groupstart   nil
     :font         (make-font :modern :courier\ new 28 '(:bold))
     :set-value-fn 'icon-edit-add-button-set-fn))
   (setf *node-edit-del-button*
    (add-icon-edit-tab-widget
     :widget       'button
     :title        "-"
     :box          (make-box 35 125 60 145)
     :3d-border    t
     :tabstop      t
     :groupstart   nil
     :font         (make-font :modern :courier\ new 28 '(:bold))
     :set-value-fn 'icon-edit-del-button-set-fn))
   (setf *node-edit-no-value-button*
    (add-icon-edit-tab-widget
     :widget       'radio-button
     :title        "No-~Value"
     :value        nil
     :box          (make-box 65 45 145 65)
     :3d-border    t
     :tabstop      t
     :groupstart   t
     :cluster      :cluster-1
     :set-value-fn 'icon-edit-no-value-button-set-fn))
   (setf *node-edit-node-button*
    (add-icon-edit-tab-widget
     :widget       'radio-button
     :title        "~Node"
     :value        nil
     :box          (make-box 65 80 125 100)
     :3d-border    t
     :tabstop      t
     :groupstart   t
     :cluster      :cluster-1
     :set-value-fn 'icon-edit-node-button-set-fn))
   (setf *node-edit-node-list*
    (add-icon-edit-tab-widget
     :widget       'combo-box
     :value        nil
     :box          (make-box 130 77 230 177)
     :tabstop      t
     :groupstart   nil
     :range        nil
     :key          #'(lambda (x)
                       (format nil "~2@a  ~a"
                               (first x)
                               (icon-class-to-name (type-of (rest x)))))
     :set-value-fn 'icon-edit-node-list-set-fn))
   (setf *node-edit-constant-button*
    (add-icon-edit-tab-widget
     :widget       'radio-button
     :title        "~Constant"
     :value        nil
     :box          (make-box 65 115 145 135)
     :3d-border    t
     :tabstop      t
     :groupstart   t
     :cluster      :cluster-1
     :set-value-fn 'icon-edit-constant-button-set-fn))
   (setf *node-edit-constant-text*
    (add-icon-edit-tab-widget
     :widget       'lisp-text
     :value        nil
     :box          (make-box 150 114 230 136)
     :3d-border    t
     :tabstop      t
     :groupstart   nil
     :set-value-fn 'icon-edit-constant-text-set-fn))))

;;; Setting up the :node tab page items and values.
(defmethod ICON-EDIT-SETUP-PAGE ((icon iconic-node) (id (eql :node))
                                 &optional socket)
  (let* ((icon-doc (icon-doc icon))
         (icon-list (icon-list icon-doc)))
    ;; Save the old icon values.
    (setf *node-saved-arguments* (copy-list (arguments icon)))
    ;; Set the values for this dialog.
    (setf *node-edit-values*
          (nconc
           (mapcar #'(lambda (x)
                       (if (iconp x)
                         (cons x +no-value+)
                         (cons nil x)))
                   (arguments icon))
           (if (list-args-p icon)
             (list (cons nil +no-value+))
             nil)))
    (let ((count -1))
      (setf (dialog-item-range *node-edit-argument-list*)
            (mapcar #'(lambda (x) (incf count))
                    (if (list-args-p icon)
                      (cons nil (arguments icon))
                      (arguments icon))))
      (setf count 0)
      (setf (dialog-item-range *node-edit-node-list*)
            (mapcar #'(lambda (x) (cons (incf count) x)) icon-list)))
    (set-dialog-item-available-p *node-edit-add-button*       nil)
    (set-dialog-item-available-p *node-edit-del-button*       nil)
    (setf (dialog-item-value     *node-edit-no-value-button*) nil)
    (set-dialog-item-available-p *node-edit-no-value-button*  nil)
    (setf (dialog-item-value     *node-edit-node-button*    ) nil)
    (set-dialog-item-available-p *node-edit-node-button*      nil)
    (setf (dialog-item-value     *node-edit-node-list*      ) nil)
    (set-dialog-item-available-p *node-edit-node-list*        nil)
    (setf (dialog-item-value     *node-edit-constant-button*) nil)
    (set-dialog-item-available-p *node-edit-constant-button*  nil)
    (setf (dialog-item-value     *node-edit-constant-text*  ) +no-value+)
    (set-dialog-item-available-p *node-edit-constant-text*    nil)
    ;; This is done with nil first to insure the set-fn call.
    (setf (dialog-item-value *node-edit-argument-list*) nil)
    (setf (dialog-item-value *node-edit-argument-list*)
          (or (and socket
                   (integerp socket)
                   (or (fixed-args-p icon) (< socket (graphic-args-num icon)))
                   socket)
              0))
    (let ((count (length icon-list)))
      (dolist (icon (reverse icon-list))
        (setf (icon-temp-string icon) count)
        (decf count)))
    (do-redraws icon-doc)))

;;; Done with the :node tab page.
(defmethod ICON-EDIT-DONE-PAGE ((icon iconic-node) (id (eql :node)) result)
  (with-delayed-redraw ((icon-doc icon))
    (if result
      ;; OK - remove redundant +no-value+'s from arguments.
      (when (list-args-p icon)
        (do ((i 0))
            ((>= i (length (arguments icon))))
          (if (eql (getarg0 icon i) +no-value+)
            (delarg0 icon i)
            (incf i))))
      ;; Cancel - restore arguments.
      (setf (arguments icon) *node-saved-arguments*)))
  ;; Clear up the mess.
  (dolist (icon (icon-list (icon-doc icon)))
    (setf (icon-temp-string icon) nil))
  (setf *node-edit-values*     nil)
  (setf *node-saved-arguments* nil))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N - E D I T  dialog message handlers                             ;;;
;;;-------------------------------------------------------------------------;;;
;;; Controlling the dialog's behavior with the set-fn's.

(defun ICON-EDIT-ARGUMENT-LIST-SET-FN (list newval oldval)
  (let ((avail (if newval t nil)))
    (set-dialog-item-available-p *node-edit-no-value-button* avail)
    (set-dialog-item-available-p *node-edit-node-button*     avail)
    (set-dialog-item-available-p *node-edit-constant-button* avail)
    (set-dialog-item-available-p *node-edit-node-list*       avail)
    (set-dialog-item-available-p *node-edit-constant-text*   avail)
    (set-dialog-item-available-p *node-edit-del-button*      avail)
    (set-dialog-item-available-p *node-edit-add-button*
                                 (if (list-args-p *icon-edit-icon*)
                                   avail nil)))
  (when newval (icon-edit-arg-to-dialog newval))
  t)

(defun ICON-EDIT-ADD-BUTTON-SET-FN (button newval oldval)
  (when (and *icon-edit-dialog-running* newval)
    (icon-edit-addarg))
  nil)

(defun ICON-EDIT-DEL-BUTTON-SET-FN (button newval oldval)
  (when (and *icon-edit-dialog-running* newval)
    (icon-edit-delarg))
  nil)

(defun ICON-EDIT-NO-VALUE-BUTTON-SET-FN (button newval oldval)
  (when (and *icon-edit-dialog-running* newval)
    (icon-edit-setarg +no-value+))
  t)

(defun ICON-EDIT-NODE-BUTTON-SET-FN (button newval oldval)
  (when (and *icon-edit-dialog-running* newval)
    (let ((val (rest (dialog-item-value *node-edit-node-list*))))
      (when val
        (icon-edit-setarg val)))
    (set-focus-widget *node-edit-node-list*))
  t)

(defun ICON-EDIT-NODE-LIST-SET-FN (list newval oldval)
  (cond ((not *icon-edit-dialog-running*) t)
        ((icon-edit-setarg (rest newval))
         (let ((*icon-edit-dialog-running* nil))
           (setf (dialog-item-value *node-edit-node-button*) t)))
        (t nil)))

(defun ICON-EDIT-CONSTANT-BUTTON-SET-FN (button newval oldval)
  (when (and *icon-edit-dialog-running* newval)
    (icon-edit-setarg (dialog-item-value *node-edit-constant-text*))
    (set-focus-widget *node-edit-constant-text*)
    (set-selection *node-edit-constant-text* 0 10000))
  t)

(defun ICON-EDIT-CONSTANT-TEXT-SET-FN (text newval oldval)
  (cond ((not *icon-edit-dialog-running*) t)
        ((icon-edit-setarg newval)
         (let ((*icon-edit-dialog-running* nil))
           (setf (dialog-item-value *node-edit-constant-button*) t)))
        (t nil)))

;;; When called with no val - used for graphics updating.
(defun ICON-EDIT-SETARG (&optional (val nil val-p))
  (let ((argnum (dialog-item-value *node-edit-argument-list*)))
    (if (or (not val-p) (check-arg *icon-edit-icon* argnum val))
      (progn
        (with-delayed-redraw ((icon-doc *icon-edit-icon*))
          (when val-p
            (if (< argnum (length (arguments *icon-edit-icon*)))
              (delarg0 *icon-edit-icon* argnum)
              (progn
                (pushlast (cons nil +no-value+) *node-edit-values*)
                (list-widget-add-item *node-edit-argument-list* (1+ argnum))))
            (setarg0 *icon-edit-icon* argnum val))
          (do-redraws (icon-doc *icon-edit-icon*)))
        (redisplay-window (icon-doc *icon-edit-icon*))
        (when val-p
          (if (iconp val)
            (setf (first (nth argnum *node-edit-values*)) val)
            (setf (rest  (nth argnum *node-edit-values*)) val)))
        t)
      nil)))

(defun ICON-EDIT-ADDARG ()
  (let ((argnum (dialog-item-value *node-edit-argument-list*)))
    (when (list-args-p *icon-edit-icon*)
      (setarg0 *icon-edit-icon* argnum +no-value+)
      (if (zerop argnum)
        (setf *node-edit-values*
              (cons (cons nil +no-value+) *node-edit-values*))
        (setf (rest (nthcdr (1- argnum) *node-edit-values*))
              (cons (cons nil +no-value+) (nthcdr argnum *node-edit-values*))))
      (list-widget-add-item
         *node-edit-argument-list*
         (length (dialog-item-range *node-edit-argument-list*)))
      (icon-edit-setarg)
      (icon-edit-arg-to-dialog argnum))))

(defun ICON-EDIT-DELARG ()
  (let ((argnum (dialog-item-value *node-edit-argument-list*)))
    (when (< argnum (length (arguments *icon-edit-icon*)))
      (delarg0 *icon-edit-icon* argnum)
      (when (list-args-p *icon-edit-icon*)
        (if (zerop argnum)
          (setf *node-edit-values* (rest *node-edit-values*))
          (setf (rest (nthcdr (1- argnum) *node-edit-values*))
                (nthcdr (1+ argnum) *node-edit-values*)))
        (list-widget-remove-item
         *node-edit-argument-list*
         (1- (length (dialog-item-range *node-edit-argument-list*)))))
      (icon-edit-setarg)
      (icon-edit-arg-to-dialog argnum))))

(defun ICON-EDIT-ARG-TO-DIALOG (arg)
  (let* ((*icon-edit-dialog-running* nil)
         (icon-arg-p (< arg (length (arguments *icon-edit-icon*))))
         (icon-arg (if icon-arg-p (getarg0 *icon-edit-icon* arg) nil))
         (saved-arg (nth arg *node-edit-values*))
         values)
    (let ((pos (position (first saved-arg)
                         (dialog-item-range *node-edit-node-list*)
                         :key #'rest)))
      (if pos
        (list-widget-set-index *node-edit-node-list* pos)
        (setf (dialog-item-value *node-edit-node-list*) nil)))
    (setf (dialog-item-value *node-edit-constant-text*) (rest saved-arg))
    (setf values
          (cond ((not icon-arg-p)          '(nil nil nil))
                ((eql icon-arg +no-value+) '( t  nil nil))
                ((iconp icon-arg)          '(nil  t  nil))
                (t                         '(nil nil  t ))))
    (mapcar #'set-dialog-item-value
            (list *node-edit-no-value-button*
                  *node-edit-node-button*
                  *node-edit-constant-button*)
            values)))


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :ICONEDIT)

