;;;                 -*- Mode: Lisp; Package: BOOMS-APP -*-                  ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         icons.lsp                                                 ;;;
;;; Description:  BOOMS iconic nodes definition                             ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Idea by:      Omri Weisman - omri@cs.bgu.ac.il                          ;;;
;;;               Cohavit Taboch  - taboch@cs.bgu.ac.il                     ;;;
;;; Date:         28-12-96                                                  ;;;
;;; Package:      BOOMS-APP                                                 ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  I C O N S  file                                                        ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :booms-app)


;;;-------------------------------------------------------------------------;;;
;;;  I C O N  class                                                         ;;;
;;;-------------------------------------------------------------------------;;;

;;; Icon class used for representing some icon on screen.
(defclass ICON ()
  ((box          :accessor icon-box         :initarg  :box)
   (box-list                                :initform nil)
   (document     :accessor icon-doc         :initform nil :initarg :document)
   (shown        :accessor shown            :initform nil)
   (selected     :accessor selected         :initform nil)
   (redraw       :accessor redraw-info      :initform (cons nil nil))
   (old-box      :accessor icon-old-box     :initform (make-box 0 0 0 0))
   (old-box-list                            :initform nil)
   (temp-str     :reader   icon-temp-string :initform nil)
   (fg-color     :accessor icon-fg-color    :initform +icon-border-color+
                                            :initarg  :fg-color)
   (bg-color     :accessor icon-bg-color    :initform +icon-bg-color+
                                            :initarg  :bg-color)
   (sel-color    :accessor icon-sel-color   :initform +icon-select-bg-color+
                                            :initarg  :sel-color)))

;;; The iconp predicate.
(defmacro ICONP (obj)
  `(typep ,obj 'icon))

;;; Make sure an icon is created valid - that its box is normalized.
(defmethod INITIALIZE-INSTANCE ((icon icon) &rest initargs)
  (unless *loading-now*
    (if (getf initargs :box)
      (normalize-box (getf initargs :box))
      (error "Should initialize an icon with a box.")))
  (call-next-method)
  icon)

;;; Making a copy of an icon.
(defmethod MAKE-COPY ((icon icon))
  (make-icon (type-of icon)
             :pos       (icon-center-pos icon)
             :centered  t
             :fg-color  (icon-fg-color  icon)
             :bg-color  (icon-bg-color  icon)
             :sel-color (icon-sel-color icon)))

;;; This may be used for icons that require some initial user interaction,
;;; return nil to cancel icon creation.  It is defined for all objects since it
;;; might be called for a list of icons (when instanciating a named copy).
(defmethod ICON-INITIALIZE ((icon t))
  t)

;;; Default icon for drawing - to be overriden later.
(defmethod ICON-ICON ((icon icon))
  question-icon)

;;; Return icon boxes list of boxes the icon occupies.
(defmethod ICON-BOX-LIST ((icon icon))
  (list (icon-box icon)))

;;; Make it memoized.
(defmethod ICON-BOX-LIST :around ((icon icon))
  (let ((box-list (slot-value icon 'box-list)))
    (or box-list
        (let ((result (call-next-method)))
          (setf (slot-value icon 'box-list) result)
          result))))

;;; Return icon old-boxes list of boxes the icon should erase.
(defmethod ICON-OLD-BOX-LIST ((icon icon))
  (list (icon-old-box icon)))

;;; Make it memoized.
(defmethod ICON-OLD-BOX-LIST :around ((icon icon))
  (let ((old-box-list (slot-value icon 'old-box-list)))
    (or old-box-list
        (let ((result (call-next-method)))
          (setf (slot-value icon 'old-box-list) result)
          result))))

;;; Get an icon's width and height.  These methods are optimized to return
;;; constants.

(defmethod ICON-WIDTH ((icon icon))
  +icon-size+)

(defmethod ICON-HEIGHT ((icon icon))
  +icon-size+)

;;; Get icon position - top-left corner of icon.  Note that it always return
;;; the same position object.
(defmethod ICON-POS ((icon icon))
  (let ((pos #.(make-position 0 0)))
    ;; Do not use nnbox-top-left because of danger of using this function with
    ;; nnbox-top-left on the same statement.
    (nbox-top-left pos (icon-box icon))))

;;; Get the icon's center position.  Note that it always return the same
;;; position object.
(defmethod ICON-CENTER-POS ((icon icon))
  (let ((pos #.(make-position 0 0)))
    (nbox-top-left pos (icon-box icon))
    (incf (position-x pos) (round (icon-width  icon) 2))
    (incf (position-y pos) (round (icon-height icon) 2))
    pos))

;;; Handle movements.

(defmethod ICON-SET-POS ((icon icon) (pos position))
  (let ((offset #.(make-position 0 0)))
    (ncopy-position offset pos)
    (nposition- offset (icon-center-pos icon))
    (dolist (box (icon-box-list icon))
      (nbox-move box offset)))
  (set-modified (icon-doc icon)))

(defmethod ICON-MOVE-POS ((icon icon) (offset position))
  (dolist (box (icon-box-list icon))
    (nbox-move box offset))
  (set-modified (icon-doc icon)))

;;; Does soething intersect with an icon?

(defmethod ICON-INTERSECT-P ((icon icon) (pos position))
  (some #'(lambda (box) (inside-box-p pos box)) (icon-box-list icon)))

(defmethod ICON-INTERSECT-P ((icon icon) (box box))
  (some #'(lambda (box1) (box-intersect-p box box1)) (icon-box-list icon)))

;;; Show the icon.
(defmethod ICON-SHOW ((icon icon))
  (unless (shown icon)
    (setf (shown icon) t)
    (request-redraw icon :show)))

;;; Hide the icon.
(defmethod ICON-HIDE ((icon icon))
  (when (shown icon)
    (setf (shown icon) nil)
    (request-redraw icon :hide)))

;;; Redisplay an icon.
(defmethod ICON-RESHOW ((icon icon))
  (when (shown icon)
    (request-redraw icon :show)))

;;; Make the icon selected/unselected/toggle.

(defmethod ICON-SELECT ((icon icon))
  (unless (selected icon)
    (setf (selected icon) t)
    (when (shown icon)
      (request-redraw icon :show))))

(defmethod ICON-UNSELECT ((icon icon))
  (when (selected icon)
    (setf (selected icon) nil)
    (when (shown icon)
      (request-redraw icon :show))))

(defmethod ICON-SELECT-TOGGLE ((icon icon))
  (setf (selected icon) (not (selected icon)))
  (when (shown icon)
    (request-redraw icon :show))
  (selected icon))

;;; Move the icon to a new position.
(defmethod ICON-MOVE ((icon icon) (newpos position))
  (if (shown icon)
    (progn
      (icon-hide icon)
      (icon-set-pos icon newpos)
      (icon-show icon))
    (icon-set-pos icon newpos)))

;;; Move the icon to by an offset.
(defmethod ICON-MOVE-RELATIVE ((icon icon) (offset position))
  (if (shown icon)
    (progn
      (icon-hide icon)
      (icon-move-pos icon offset)
      (icon-show icon))
    (icon-move-pos icon offset)))

;;; Snap icons to grid.
(defmethod ICON-TO-GRID ((icon icon))
  (let* ((offset #.(make-position 0 0))
         (icon-pos (icon-center-pos icon))
         (x (position-x icon-pos))
         (y (position-y icon-pos)))
    (setf x (- (* (round x +grid-size-x+) +grid-size-x+) x))
    (setf y (- (* (round y +grid-size-y+) +grid-size-y+) y))
    (unless (and (zerop x) (zerop y))
      (nmake-position offset x y)
      (icon-move-relative icon offset))))

;;; Reset redraw-info.
(defmethod RESET-REDRAW-INFO ((icon icon))
  (let ((redraw-info (redraw-info icon)))
    (setf (first redraw-info) nil)
    (setf (rest  redraw-info) nil)))

;;; Is there any pending redraw?
(defmethod REDRAW-P ((icon icon))
  (let ((redraw-info (redraw-info icon)))
    (not (and (null (first redraw-info)) (null (rest redraw-info))))))

;;; Request a redraw action.
(defmethod REQUEST-REDRAW ((icon icon) action)
  (unless (redraw-p icon)
    (request-icon-redraw (icon-doc icon) icon))
  (let ((redraw-info (redraw-info icon)))
    (case action
      (:show (setf (rest redraw-info) t))
      (:hide (unless (first redraw-info)
               (setf (first redraw-info) t)
               (copy-boxes-to-old-boxes icon))))))

;;; Copy the icon current boxes to its old-boxes for erase region save.
(defmethod COPY-BOXES-TO-OLD-BOXES ((icon icon))
  (ncopy-box (icon-old-box icon) (icon-box icon)))

;;; Setting the temp-str requires a redraw.
(defmethod (setf ICON-TEMP-STRING) (newval (icon icon))
  (setf (slot-value icon 'temp-str) (and newval (capitalize-object newval)))
  (icon-reshow icon))

;;; Do the icon redraw commands.
(defmethod DO-ICON-REDRAWS ((icon icon))
  (do-icon-erase-old icon)
  (do-icon-draw icon))

;;; Do the icon draw request.
(defmethod DO-ICON-DRAW ((icon icon))
  (let ((redraw-info (redraw-info icon)))
    (when (rest redraw-info)
      (icon-draw icon)
      (setf (rest redraw-info) nil))))

;;; Do the icon old-boxes erasing request.
(defmethod DO-ICON-ERASE-OLD ((icon icon))
  (let ((redraw-info (redraw-info icon)))
    (when (first redraw-info)
      (icon-erase-old icon)
      (setf (first redraw-info) nil))))

;;; Draw the icon frame.
(defmethod ICON-DRAW-FRAME ((icon icon))
  (dolist (box (icon-box-list icon))
    (draw-box (icon-doc icon) box)))

;;; Draw the icon border - like the draw frame but bound by the boxes.
(defmethod ICON-DRAW-BORDER ((icon icon))
  (dolist (box (icon-box-list icon))
    (with-clipping ((icon-doc icon) box)
      (draw-box (icon-doc icon) box))))

;;; Draw the icon background.
(defmethod ICON-DRAW-BACKGROUND ((icon icon))
  (with-foreground-color
      ((icon-doc icon)
       (if (selected icon) (icon-sel-color icon) (icon-bg-color icon)))
    (dolist (box (icon-box-list icon))
      (fill-box (icon-doc icon) box)))
  (with-foreground-color ((icon-doc icon) (icon-fg-color icon))
    (with-line-width ((icon-doc icon) +border-line-width+)
      (icon-draw-border icon))))

;;; Draw the icon's Windows icon.
(defmethod ICON-DRAW-ICON ((icon icon))
  (let ((pos #.(make-position 0 0)))
    (nmake-position pos +icon-margin+ +icon-margin+)
    (draw-icon (icon-doc icon)
               (icon-icon icon)
               (nposition+ pos (icon-pos icon)))))

;;; Draw the temp string on the icon.
(defmethod ICON-DRAW-STRING ((icon icon))
  (let ((str (icon-temp-string icon))
        (doc (icon-doc icon))
        (box (icon-box icon))
        (pos #.(make-position +border-line-width+ +border-line-width+))
        (drw #.(make-box 0 0 0 0)))
    (with-font (doc +icon-temp-text-font+)
      (ncopy-box drw box)
      (nbox-move drw pos)
      (set-box-right
       drw (+ (box-left drw) (stream-string-width doc str) +icon-text-margin+))
      (set-box-bottom
       drw (+ (box-top drw) (line-height doc) +icon-text-margin+))
      (with-clipping (doc box)
        (with-foreground-color (doc +icon-text-bg-color+)
          (fill-box doc drw))
        (with-foreground-color (doc +icon-text-fg-color+)
          (with-background-color (doc +icon-text-bg-color+)
            (draw-string-in-box doc str 0 (length str) drw :center :center nil)
            (draw-box doc drw)))))))

;;; Draw the icon.
(defmethod ICON-DRAW ((icon icon))
  (icon-draw-background icon)
  (icon-draw-icon icon)
  (when (icon-temp-string icon) (icon-draw-string icon)))

;;; Don't draw the icon when hidden.
(defmethod ICON-DRAW :around ((icon icon))
  (when (shown icon) (call-next-method)))

;;; Erase the saved old box.
(defmethod ICON-ERASE-OLD ((icon icon))
  (dolist (box (icon-old-box-list icon))
    (erase-contents-box (icon-doc icon) box)))

;;; Setting these attributes means the document is modified.

(defmethod (setf ICON-FG-COLOR)  :before (newval (icon icon))
  (set-modified (icon-doc icon)))

(defmethod (setf ICON-BG-COLOR)  :before (newval (icon icon))
  (set-modified (icon-doc icon)))

(defmethod (setf ICON-SEL-COLOR) :before (newval (icon icon))
  (set-modified (icon-doc icon)))

;;; Return the icon-style for some icon (nothing for non-icons).

(defmethod ICON-STYLE ((obj t)) nil)

(defmethod ICON-STYLE ((icon icon))
  (list* :fg-color  (icon-fg-color  icon)
         :bg-color  (icon-bg-color  icon)
         :sel-color (icon-sel-color icon)
         (call-next-method)))

;;; Set the icon-style from some list (do nothing for non-nodes).

(defmethod (setf ICON-STYLE) (style (obj t)) t)

(defmethod (setf ICON-STYLE) (style (icon icon))
  (when (getf style :fg-color)
    (setf (icon-fg-color  icon) (getf style :fg-color)))
  (when (getf style :bg-color)
    (setf (icon-bg-color  icon) (getf style :bg-color)))
  (when (getf style :sel-color)
    (setf (icon-sel-color icon) (getf style :sel-color)))
  (call-next-method)
  (icon-reshow icon))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N - L I S T  handling                                            ;;;
;;;-------------------------------------------------------------------------;;;

;;; This function returns the center position of an icon list.
(defun ICONS-CENTER-POSITION (icon-list &optional (center-of-centers t))
  (let ((pos #.(make-position 0 0))
        (len (length icon-list)))
    (nmake-position pos 0 0)
    (if center-of-centers
      (dolist (icon icon-list) (nposition+ pos (icon-center-pos icon)))
      (dolist (icon icon-list) (nposition+ pos (icon-pos icon))))
    (setf (position-x pos) (round (position-x pos) len))
    (setf (position-y pos) (round (position-y pos) len))
    pos))

;;; This will move a list of icons so its center will be in some position.
;;; This always work on the center of centers.  New-pos is copied first because
;;; it is very possible that it is the result of icon[s]-center-pos and here we
;;; use it also (this is why the let* and not simple let).
(defun (setf ICONS-CENTER-POSITION) (newpos icon-list)
  (let* ((moveby (ncopy-position #.(make-position 0 0) newpos))
         (center (icons-center-position icon-list)))
    (nposition- moveby center)
    (dolist (icon icon-list)
      (icon-move-relative icon moveby))))


;;;-------------------------------------------------------------------------;;;
;;;  S O C K E T S - I C O N  class                                         ;;;
;;;-------------------------------------------------------------------------;;;
;;; This supplies the sockets functionality.  It should be mixed with the node
;;; class at some stage.

(defclass SOCKETS-ICON (icon)
  ((sockets-box     :accessor sockets-box     :initform (make-box 0 0 0 0))
   (sockets-margin  :accessor sockets-margin  :initform 0)
   (sockets-old-box :accessor sockets-old-box :initform (make-box 0 0 0 0))))

;;; When making a sockets-icon - adjust the attached sockets.
(defmethod INITIALIZE-INSTANCE ((icon sockets-icon) &rest initargs)
  (call-next-method)
  (unless *loading-now*
    (set-box-bottom (icon-box icon)
                    (- (box-bottom (icon-box icon)) +socket-margin+))
    (adjust-sockets-box icon))
  icon)

;;; The number of nodes visible - has an extra in case of a list of arguments.
(defmethod GRAPHIC-ARGS-NUM ((icon sockets-icon))
  (if (list-args-p icon)
    (1+ (length (arguments icon)))
    (args-num icon)))

;;; Find weather a position is inside a socket.  Return its number or nil.
(defmethod ICON-FIND-SOCKETS ((icon sockets-icon) (pos position))
  (and (<= (+ (box-top (sockets-box icon)) +socket-margin+)
           (position-y pos)
           (+ (box-top (sockets-box icon)) +socket-total-size+))
       (<= (+ (box-left (sockets-box icon)) +socket-margin+
              (sockets-margin icon))
           (position-x pos))
       (let ((x (- (position-x pos)
                   (box-left (sockets-box icon))
                   +socket-margin+ (sockets-margin icon))))
         (and (>= x 0)
              (multiple-value-bind (div mod) (floor x +socket-total-size+)
                (and (< mod +socket-size+)
                     (< div (graphic-args-num icon))
                     div))))))

;;; Return the center position of a socket.  Note that it always return the
;;; same position object.
(defmethod SOCKETS-CENTER ((icon sockets-icon) (socket integer)
                           &optional use-old-box)
  (let ((box (if (and use-old-box (redraw-p icon) (first (redraw-info icon)))
               (sockets-old-box icon)
               (sockets-box icon)))
        (pos #.(make-position 0 0)))
    (nmake-position pos
                    (+ (box-left box) (sockets-margin icon)
                       (* socket +socket-total-size+)
                       #.(+ +socket-margin+ (round +socket-size+ 2)))
                    (+ (box-top box)
                       #.(+ +socket-margin+ (round +socket-size+ 2))))))

;;; Add the sockets box to the box list.
(defmethod ICON-BOX-LIST ((icon sockets-icon))
  (cons (sockets-box icon) (call-next-method)))

;;; Add the old sockets box to the old box list.
(defmethod ICON-OLD-BOX-LIST ((icon sockets-icon))
  (cons (sockets-old-box icon) (call-next-method)))

;;; The height of sockets-icon includes the sockets height.
(defmethod ICON-HEIGHT ((icon sockets-icon))
  #.(+ +icon-size+ +socket-total-size+ +socket-margin+))

;;; Adjust the sockets-box according to current arguments.
(defmethod ADJUST-SOCKETS-BOX ((icon sockets-icon))
  (when (args-num icon)
    (let* ((box    (icon-box icon))
           (margin (round (+ +socket-margin+
                             (* +socket-total-size+ (graphic-args-num icon))
                             (- (box-width box))) 2))
           (shown  (shown icon)))
      (when shown (icon-hide icon))
      (if (minusp margin)
        (progn
          (setf (sockets-margin icon) (- margin))
          (setf margin 0))
        (setf (sockets-margin icon) 0))
      (nmake-box (sockets-box icon)
                 (- (box-left box) margin)
                 (box-bottom box)
                 (+ (box-right box) margin)
                 (+ (box-bottom box) +socket-total-size+ +socket-margin+))
      (when shown (icon-show icon)))))

;;; Copy the icon current boxes to its old-boxes includes the sockets-old-box.
(defmethod COPY-BOXES-TO-OLD-BOXES ((icon sockets-icon))
  (call-next-method)
  (ncopy-box (sockets-old-box icon) (sockets-box icon)))

;;; Draw the icon frame with the sockets box.
(defmethod ICON-DRAW-FRAME ((icon sockets-icon))
  (let ((icon-box    (icon-box icon))
        (sockets-box (sockets-box icon))
        (points
         ;; An eight-point vector for the frame.
         #.(vector
            (make-position 0 0) (make-position 0 0) (make-position 0 0)
            (make-position 0 0) (make-position 0 0) (make-position 0 0)
            (make-position 0 0) (make-position 0 0))))
    (nbox-bottom-right (aref points 0) icon-box)
    (nbox-top-right    (aref points 1) icon-box)
    (nbox-top-left     (aref points 2) icon-box)
    (nbox-bottom-left  (aref points 3) icon-box)
    (nbox-top-left     (aref points 4) sockets-box)
    (nbox-bottom-left  (aref points 5) sockets-box)
    (nbox-bottom-right (aref points 6) sockets-box)
    (nbox-top-right    (aref points 7) sockets-box)
    (draw-polygon (icon-doc icon) points)))

;;; Draw the icon-border with the sockets box, use the bounding boxes.
(defmethod ICON-DRAW-BORDER ((icon sockets-icon))
  (let ((doc         (icon-doc icon))
        (icon-box    (icon-box icon))
        (sockets-box (sockets-box icon))
        (points1 #.(vector (make-position 0 0) (make-position 0 0)
                           (make-position 0 0) (make-position 0 0)))
        (points2 #.(vector (make-position 0 0) (make-position 0 0)
                           (make-position 0 0) (make-position 0 0)
                           (make-position 0 0) (make-position 0 0))))
    (nbox-bottom-right (aref points1 0) icon-box)
    (nbox-top-right    (aref points1 1) icon-box)
    (nbox-top-left     (aref points1 2) icon-box)
    (nbox-bottom-left  (aref points1 3) icon-box)
    (with-clipping (doc icon-box)
      (draw-polyline doc points1))
    (nbox-bottom-left  (aref points2 0) icon-box)
    (nbox-top-left     (aref points2 1) sockets-box)
    (nbox-bottom-left  (aref points2 2) sockets-box)
    (nbox-bottom-right (aref points2 3) sockets-box)
    (nbox-top-right    (aref points2 4) sockets-box)
    (nbox-bottom-right (aref points2 5) icon-box)
    (with-clipping (doc sockets-box)
      (draw-polyline doc points2))))

;;; Draw the sockets.
(defmethod ICON-DRAW ((icon sockets-icon))
  (call-next-method)
  (let ((box #.(make-box 0 0 0 0)))
    (nmake-box-relative box
                        (+ (box-left (sockets-box icon)) +socket-margin+
                           (sockets-margin icon))
                        (+ (box-top  (sockets-box icon)) +socket-margin+)
                        +socket-size+ +socket-size+)
    (dotimes (socket (graphic-args-num icon))
      (with-foreground-color
          ((icon-doc icon)
           (cond
             ((= socket (length (arguments icon)))   +socket-holder-color+)
             ((eql (getarg0 icon socket) +no-value+) +socket-empty-color+)
             ((typep (getarg0 icon socket) 'icon)    +socket-full-color+)
             (t                                      +socket-constant-color+)))
        (fill-box (icon-doc icon) box))
      (nbox-move box #.(make-position +socket-total-size+ 0)))))


;;;-------------------------------------------------------------------------;;;
;;;  T O P - S O C K E T  class                                             ;;;
;;;-------------------------------------------------------------------------;;;
;;; This supplies the top-socket functionality.

(defclass TOP-SOCKET-ICON (icon) ())

;;; When making an top-socket-icon node - adjust the icon box.
(defmethod INITIALIZE-INSTANCE ((icon top-socket-icon) &rest initargs)
  (call-next-method)
  (unless *loading-now*
    (set-box-bottom (icon-box icon)
                    (+ (box-bottom (icon-box icon)) +socket-total-size+)))
  icon)

;;; The height of top-socket-icon includes the top-socket height.
(defmethod ICON-HEIGHT ((icon top-socket-icon))
  #.(+ +icon-size+ +socket-total-size+))

;;; Return the actual icon pos.  Note that it always return the same position
;;; object.  Note that it always return the same position object.
(defmethod ICON-POS ((icon top-socket-icon))
  (let ((pos #.(make-position 0 0)))
    (nmake-position pos
                    (box-left (icon-box icon))
                    (+ (box-top (icon-box icon)) +socket-total-size+))))

;;; Find weather a position is inside the top socket.  Return :top/nil.
(defmethod ICON-FIND-TOP-SOCKET ((icon top-socket-icon) (pos position))
  (and (<= (+ (box-top (icon-box icon)) +socket-margin+)
           (position-y pos)
           (+ (box-top (icon-box icon)) +socket-total-size+))
       (<= (+ (box-left (icon-box icon)) +top-side-margin+)
           (position-x pos)
           (- (box-right (icon-box icon)) +top-side-margin+))
       :top))

;;; Return the center position of top socket.  Note that it always return the
;;; same position object.
(defmethod TOP-SOCKET-CENTER ((icon top-socket-icon) &optional use-old-box)
  (let ((box (if (and use-old-box (redraw-p icon) (first (redraw-info icon)))
               (icon-old-box icon)
               (icon-box icon)))
        (pos #.(make-position 0 0)))
    (nmake-position pos
                    (+ (box-left box)
                       #.(+ +top-side-margin+ (round +socket-size+ 2)))
                    (+ (box-top box)
                       #.(+ +socket-margin+ (round +socket-size+ 2))))))

;;; Draw the top socket.
(defmethod ICON-DRAW ((icon top-socket-icon))
  (call-next-method)
  (with-foreground-color ((icon-doc icon) +socket-top-color+)
    (fill-box (icon-doc icon)
              (nmake-box-relative
               #.(make-box 0 0 0 0)
               (+ (box-left (icon-box icon)) +top-side-margin+)
               (+ (box-top  (icon-box icon)) +socket-margin+)
               +socket-size+ +socket-size+))))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N I C - N O D E  class                                           ;;;
;;;-------------------------------------------------------------------------;;;
;;; Iconic-node is the class of node icons, actually a mixin - it should serve
;;; only as a base class for the deficon macro that will combine it as well as
;;; node into a functional class.

;;; Order of superclasses is important.
(defclass ICONIC-NODE (sockets-icon top-socket-icon node)
  ((lines-shown :accessor lines-shown :initform nil)))

;;; After setting/deleting an argument[s] - adjust sockets in case of a change
;;; in the arguments number, and redisplay.

(defmethod SETARG0 ((n iconic-node) (pos integer) (what t))
  (let ((len (length (arguments n))) (shown (shown n)))
    (icon-lines-hide n nil)
    (when shown (icon-hide n))
    (prog1 (call-next-method)
      (unless (= len (length (arguments n))) (adjust-sockets-box n))
      (when shown (icon-show n)))))

(defmethod DELARG0 ((n iconic-node) (pos integer))
  (let ((len (length (arguments n))) (shown (shown n)))
    (icon-lines-hide n nil)
    (when shown (icon-hide n))
    (prog1 (call-next-method)
      (unless (= len (length (arguments n))) (adjust-sockets-box n))
      (when shown (icon-show n)))))

(defmethod (setf ARGUMENTS) ((new-args list) (n iconic-node))
  (let ((len (length (arguments n))) (shown (shown n)))
    (icon-lines-hide n nil)
    (when shown (icon-hide n))
    (prog1 (call-next-method)
      (unless (= len (length (arguments n))) (adjust-sockets-box n))
      (when shown (icon-show n))
      (set-modified (icon-doc n)))))

;;; Describing an iconic-node with a string.
(defmethod ICON-DESCRIBE ((icon iconic-node) &optional socket)
  (format nil "~:(~a~) ~:[not-evaluated~*~;= ~a~]~:[~;      [~:*~a]~]"
          (icon-describe-name icon)
          (evaluated-p icon) (and (evaluated-p icon) (evaluate icon))
          (and socket (icon-describe-socket icon socket))))

;;; This is used to give icon-classes some value [symbol] which is better for
;;; display (like in the classes menu / outline) - this might not be the class
;;; symbol name.  If the class-name is not a symbol, then simply return it.
(defun ICON-CLASS-TO-NAME (class-symbol)
  (or (and (not (symbolp class-symbol)) class-symbol)
      (get class-symbol :booms-name nil) class-symbol))

;;; Macro for simple booms-name declarations.
(defmacro SET-BOOMS-NAME (class name)
  `(put ',class :booms-name ,name))

;;; The name of an icon - for the description.
(defmethod ICON-DESCRIBE-NAME ((icon iconic-node))
  (icon-class-to-name (type-of icon)))

;;; Describing an iconic-node's socket.
(defmethod ICON-DESCRIBE-SOCKET ((icon iconic-node) socket)
  (cond
    ((eql socket :top)
     (format nil "output socket : ~a = ~a"
             (type-out icon)
             (if (evaluated-p icon) (evaluate icon) "not-evaluated")))
    ((and (integerp socket) (list-args-p icon))
     (format nil "~:R ~a socket : ~a = ~a"
             (1+ socket) (arg-names icon) (second (type-in icon))
             (cond ((= socket (length (arguments icon))) "place-holder")
                   ((iconp (getarg0 icon socket))
                    (icon-describe (getarg0 icon socket)))
                   (t (getarg0 icon socket)))))
    ((and (integerp socket) (fixed-args-p icon))
     (format nil "~a socket : ~a = ~a"
             (nth socket (arg-names icon)) (nth (1+ socket) (type-in icon))
             (if (iconp (getarg0 icon socket))
               (icon-describe (getarg0 icon socket))
               (getarg0 icon socket))))))

;;; Evaluating an iconic-node.
(defmethod ICON-EVAL ((icon iconic-node))
  (unless (evaluated-p icon) (say "Evaluating..."))
  (let ((val (evaluate icon)))
    (if (evaluated-p icon)
      (progn (say "Evaluated to ~a" (evaluate icon)) (evaluate icon))
      (progn (say! "Not evaluated!") nil))))

;;; The height of iconic-nodes.
(defmethod ICON-HEIGHT ((icon iconic-node))
  #.(+ +icon-size+ +socket-total-size+ +socket-margin+ +socket-total-size+))

;;; Get the default socket number for a given value.  For a list-args - give
;;; the last one (or nil in case of type mismatch) and for a fixed-args - find
;;; the first with the matching type.
(defmethod DEFAULT-SOCKET-NUMBER ((icon iconic-node) val)
  (cond
    ((list-args-p icon)
     (if (check-arg icon (length (arguments icon)) val)
       (length (arguments icon))
       nil))
    ((fixed-args-p icon)
     (dotimes (pos (first (type-in icon)))
       (when (check-arg icon pos val t)
         (return-from default-socket-number pos)))
     nil)
    (t nil)))

;;; Return the socket a position is on - :top is the top one, a number
;;; (starting from 0) is a socket at the bottom and nil is none.
(defmethod ICON-FIND-SOCKET ((icon iconic-node) (pos position))
  (or (icon-find-top-socket icon pos) (icon-find-sockets icon pos)))

;;; Return the center position of a socket.  Note that it always return the
;;; same position object (for each method - sockets or top-socket).
(defmethod SOCKET-CENTER ((icon iconic-node) socket &optional use-old-box)
  (if (integerp socket)
    (sockets-center icon socket use-old-box)
    (top-socket-center icon use-old-box)))

;;; Draw/Erase the connections leading down.

(defmethod ICON-LINES-SHOW ((icon iconic-node))
  (unless (lines-shown icon)
    (with-paint-operation ((icon-doc icon) po-invert)
      (icon-draw-connections icon))
    (setf (lines-shown icon) t)))

(defmethod ICON-LINES-HIDE ((icon iconic-node) &optional (use-old-box t))
  (when (lines-shown icon)
    (with-paint-operation ((icon-doc icon) po-invert)
      (icon-draw-connections icon use-old-box))
    (setf (lines-shown icon) nil)))

;;; Draw the connections leading down.
(defmethod ICON-DRAW-CONNECTIONS ((icon iconic-node) &optional use-old-box)
  (when (shown icon)
    (let ((i 0))
      (dolist (arg (arguments icon))
        (when (and (typep arg 'icon) (shown arg))
          (draw-line (icon-doc icon)
                     (socket-center icon i use-old-box)
                     (socket-center arg :top use-old-box)))
        (incf i)))))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N - C O N S T R U C T O R S                                      ;;;
;;;-------------------------------------------------------------------------;;;

;;; Construct an icon, given its class name.
(defmethod MAKE-ICON ((icon-class symbol)
                      &rest initargs
                      &key (pos nil) (centered nil)
                      &allow-other-keys)
  (when (and pos (not (getf initargs :box)))
    (setf (getf initargs :box)
          (if centered
            (make-box 0 0 +icon-size+ +icon-size+)
            (make-box-relative-from-corner pos +icon-size+ +icon-size+))))
  (remf initargs :pos)
  (remf initargs :centered)
  (let ((newicon (apply #'make-instance icon-class initargs)))
    (when (and pos centered)
      (icon-move newicon pos))
    newicon))


;;;-------------------------------------------------------------------------;;;
;;;  D E F I C O N  macro                                                   ;;;
;;;-------------------------------------------------------------------------;;;

;;; Deficon is used to define new icons, the difference here is that
;;; iconic-node is insured to be a superclass.
(defmacro DEFICON (name arguments ancestors slots &rest methods)
  (setf ancestors (insure-superclass ancestors 'iconic-node))
  `(defnode ,name ,arguments ,ancestors ,slots ,@methods))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N  menus                                                         ;;;
;;;-------------------------------------------------------------------------;;;

;;; This will be bound to the current icon for the popped menu.
(defvar *CURRENT-MENU-ICON* nil)

;;; Add a menu item to an icon menu.
(defun ADD-ICON-MENU-ITEM (menu action title helpstr)
  (add-to-menu menu (make-menu-item :title       title
                                    :value       action
                                    :help-string helpstr)))

;;; The function that will call a menu-item function on the current icon.
(defun FUNCALL-ICON-MENU-FUNC (menu menu-item window)
  (funcall (menu-item-value menu-item) *current-menu-icon*))

;;; Create an icon menu - put it on a global list to close when BOOMS exits.
(defun MAKE-DEFAULT-ICON-MENU ()
  (let ((menu (open-menu nil 'my-pull-down-menu *screen*
                         :selection-function 'funcall-icon-menu-func)))
    (push menu *misc-menus*)
    (add-icon-menu-item menu 'icon-eval "E~vluate" "Evaluate this icon")
    (add-icon-menu-item menu 'icon-edit "~Edit"    "Edit this icon")
    (add-to-menu menu menu-separator)
    menu))

;;; Popping an icon menu.
(defmethod POP-UP-ICON-MENU ((icon icon) (win window) (pos position))
  (let ((*current-menu-icon* icon))
    (pop-up-menu (icon-menu icon) win pos nil nil :right)))

;;; The default icon variable & menu.

(defvar *DEFAULT-ICON-MENU* nil)

(defmethod ICON-MENU ((icon icon))
  *default-icon-menu*)

;;; Create the default menu.
(defun INITIALIZE-ICON-MENU  ()
  (setf *default-icon-menu* (make-default-icon-menu)))


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :ICONS)

