;;;                 -*- Mode: Lisp; Package: BOOMS-APP -*-                  ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         leaves.lsp                                                ;;;
;;; Description:  BOOMS iconic leaves definition                            ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Date:         29-12-96                                                  ;;;
;;; Package:      BOOMS-APP                                                 ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  L E A V E S  file                                                      ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :booms-app)


;;;-------------------------------------------------------------------------;;;
;;;  I C O N I C - L E A F  class                                           ;;;
;;;-------------------------------------------------------------------------;;;
;;; Iconic-leaf is the class of leaf icons, actually a mixin - it should serve
;;; only as a base class for the defleaf macro that will combine it as well as
;;; node into a functional class.  Some of the iconic-leaf functionality is
;;; copied from the iconic-node, it can be improved using some base class.

;;; Order of superclasses is important.
(defclass ICONIC-LEAF (top-socket-icon icon leaf)
  ((lines-shown :accessor lines-shown :initform nil)))

;;; Making a copy of an iconic-leaf.
(defmethod MAKE-COPY ((icon iconic-leaf))
  (make-icon (type-of icon)
             :pos       (box-top-left (icon-box icon))
             :fg-color  (icon-fg-color  icon)
             :bg-color  (icon-bg-color  icon)
             :sel-color (icon-sel-color icon)
             :type-out  (type-out icon)
             :arguments (make-copy (arguments icon))))

;;; This envelopes the leaf method for possible drawing change.
(defmethod (setf ARGUMENTS) (newval (l iconic-leaf))
  (let ((shown (shown l)))
    (icon-lines-hide l nil)
    (when shown (icon-hide l))
    (prog1 (call-next-method)
      (when shown (icon-show l))
      (set-modified (icon-doc l)))))

;;; Describing an iconic-leaf with a string.
(defmethod ICON-DESCRIBE ((icon iconic-leaf) &optional socket)
  (format nil "~:(~a~) ~a"
          (icon-describe-name icon)
          (arguments icon)))

;;; The name of an iconic-leaf - for the description.
(defmethod ICON-DESCRIBE-NAME ((icon iconic-leaf))
  (icon-class-to-name (type-of icon)))

;;; Evaluating an iconic-leaf does nothing but return the arguments (=value).
(defmethod ICON-EVAL ((icon iconic-leaf))
  (say "Leaf value is ~a" (evaluate icon))
  (evaluate icon))

;;; There are no sockets on a leaf.
(defmethod DEFAULT-SOCKET-NUMBER ((icon iconic-leaf) val)
  nil)

;;; This can only return :top now.
(defmethod ICON-FIND-SOCKET ((icon iconic-leaf) (pos position))
  (icon-find-top-socket icon pos))

;;; Return the center position of only the top socket.  socket is assumed
;;; always to be :top.
(defmethod SOCKET-CENTER ((icon iconic-leaf) socket &optional use-old-box)
  (top-socket-center icon use-old-box))

;;; Draw/Erase the connections leading down -- nothing to do with leaf icons!

(defmethod ICON-LINES-SHOW ((icon iconic-leaf))
  t)

(defmethod ICON-LINES-HIDE ((icon iconic-leaf) &optional (use-old-box t))
  t)

;;; User query - ask for value using an edit pane, set arguments first.
(defmethod ICON-INITIALIZE ((icon iconic-leaf))
  (multiple-value-bind (default pane) (leaf-init-val/pane icon)
    (if pane
        (progn
          (setf (arguments icon) default)
          (icon-edit icon nil pane t))
        t)))

;;; This is used in the above for selecting an init value and a pane for
;;; editing this value.  Return nil for not using an edit pane.
(defmethod LEAF-INIT-VAL/PANE ((icon iconic-leaf)) (values nil :leaf))


;;;-------------------------------------------------------------------------;;;
;;;  D E F L E A F  macro                                                   ;;;
;;;-------------------------------------------------------------------------;;;

;;; Defleaf is used to define new leaf subclasses.  It is much simpler than the
;;; defnode/icon macros since it should only define the leaf subclass - no
;;; arguments or methods attached.
(defmacro DEFLEAF (name ancestors slots)
  `(defclass ,name ,(insure-superclass ancestors 'iconic-leaf) ,slots))


;;;-------------------------------------------------------------------------;;;
;;;  L E A F  menu and edit page                                            ;;;
;;;-------------------------------------------------------------------------;;;

;;; The leaf menu is simply the default one, can expand this later.
(defun MAKE-LEAF-ICON-MENU ()
  (let ((menu (make-default-icon-menu)))
    menu))

;;; The default leaf variable & menu.

(defvar *LEAF-ICON-MENU* nil)

(defmethod ICON-MENU ((icon iconic-leaf))
  *leaf-icon-menu*)

;;; The leaf lisp-text widget.
(defvar *LEAF-EDIT-TEXT* nil)

;;; The leaf edit page.
(defun MAKE-ICONIC-LEAF-TAB ()
  (add-icon-edit-tab :leaf)
  (progrev
   (add-icon-edit-tab-widget
     :widget       'static-text
     :value        "~Value"
     :box          (make-box 40 30 200 50)
     :tabstop      nil
     :groupstart   t)
   (setf *leaf-edit-text*
    (add-icon-edit-tab-widget
     :widget       'lisp-text
     :value        nil
     :box          (make-box 40 60 200 90)
     :3d-border    t
     :tabstop      t
     :groupstart   nil
     :set-value-fn #'(lambda (text newval oldval)
                       (cond
                         ((not *icon-edit-dialog-running*)     t)
                         ((check-args *icon-edit-icon* newval) t)
                         (t nil)))))
   (add-icon-edit-tab-widget
     :widget       'button
     :title        "~A p p l y"
     :box          (make-box 40 100 200 130)
     :3d-border    t
     :tabstop      t
     :groupstart   nil
     :set-value-fn #'(lambda (&rest r)
                       (let ((newval (dialog-item-value *leaf-edit-text*)))
                         (setf (arguments *icon-edit-icon*) newval)
                         (do-redraws (icon-doc *icon-edit-icon*))
                         nil)))))

;;; Initialize the page values.
(defmethod ICON-EDIT-SETUP-PAGE ((icon iconic-leaf) (id (eql :leaf))
                                 &optional socket)
  (set-dialog-item-value *leaf-edit-text* (arguments icon)))

;;; Terminate the page.
(defmethod ICON-EDIT-DONE-PAGE ((icon iconic-leaf) (id (eql :leaf)) result)
  (when result
    (setf (arguments icon) (dialog-item-value *leaf-edit-text*))))

;;; Return this page for iconic-leaf.
(defmethod ICON-TAB-IDS ((x iconic-leaf))
  (list* :leaf (call-next-method)))

;;; Create the default menu.
(defun INITIALIZE-ICONIC-LEAF  ()
  (make-iconic-leaf-tab)
  (setf *leaf-icon-menu* (make-leaf-icon-menu)))


;;;-------------------------------------------------------------------------;;;
;;;  I C O N I C - C O N S T A N T  class                                   ;;;
;;;-------------------------------------------------------------------------;;;

;;; This is the base class for constant leaves.
(defleaf ICONIC-CONSTANT () ())

;;; A macro for defining new constant icon classes.
(defmacro DEFCONSTLEAF (name ancestors slots)
  `(defclass ,name ,(insure-superclass ancestors 'iconic-constant) ,slots))

;;; Cannot modify stored value.
(defmethod (setf ARGUMENTS) (newval (a iconic-constant))
  nil)

;;; Constants only have the colors page.
(defmethod ICON-TAB-IDS ((x iconic-constant))
  (list :colors))

;;; No user query for constants.
(defmethod ICON-INITIALIZE ((icon iconic-constant))
  t)


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :LEAVES)

