;;;                 -*- Mode: Lisp; Package: BOOMS-APP -*-                  ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         saveload.lsp                                              ;;;
;;; Description:  BOOMS save and load methods                               ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Idea by:      Omri Weisman - omri@cs.bgu.ac.il                          ;;;
;;;               Cohavit Taboch  - taboch@cs.bgu.ac.il                     ;;;
;;; Date:         27-12-96                                                  ;;;
;;; Package:      BOOMS-APP                                                 ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  S A V E L O A D  file                                                  ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :booms-app)


;;;-------------------------------------------------------------------------;;;
;;;  L O A D - S L O T S                                                    ;;;
;;;-------------------------------------------------------------------------;;;
;;; This is defined using the append standard method combination so we'll get
;;; the full slot list for each class according to its superclasses.

(defgeneric LOAD-SLOTS (x) (:method-combination append))

;;; Load slots of each class.

(defmethod LOAD-SLOTS append ((x t))
  nil)

(defmethod LOAD-SLOTS append ((x node))
  '(arg-names uplinks arguments))

(defmethod LOAD-SLOTS append ((x leaf))
  '(type-out))

(defmethod LOAD-SLOTS append ((x abstraction))
  '(root internal-nodes node/argss type-in))

(defmethod LOAD-SLOTS append ((x icon))
  '(box shown fg-color bg-color sel-color))

(defmethod LOAD-SLOTS append ((x sockets-icon))
  '(sockets-box sockets-margin))

(defmethod LOAD-SLOTS append ((x top-socket-icon))
  '())

(defmethod LOAD-SLOTS append ((x iconic-node))
  '())

;;; Saving a node or an icon.

(defmethod MAKE-LOAD-FORM ((icon icon))
  (make-load-form-saving-slots icon (load-slots icon)))

(defmethod MAKE-LOAD-FORM ((node node))
  (make-load-form-saving-slots node (load-slots node)))

;;; A class for non-node objects that can be saved.

(defclass SAVABLE () ())

(defmethod MAKE-LOAD-FORM ((obj savable))
  (make-load-form-saving-slots obj (load-slots obj)))


;;;-------------------------------------------------------------------------;;;
;;;  D O C - S A V E  definition                                            ;;;
;;;-------------------------------------------------------------------------;;;

;;; Saving a document.
(defmethod DOC-SAVE ((doc document))
  (with-lock *prevent-actions*
    (with-open-file (file (title doc) :direction :output)
      (if file
        (progn
          (say "Saving ~a..." (title doc))
          (fasl-write file :booms-file)
          (fasl-write file (doc-type doc))
          (doc-save-data doc file)
          (say "Done.")
          (setf (modified-p doc) nil)
          (close file))
        (my-pop-up-message-dialog *booms-app* "Booms Error"
                                  "Error saving document"
                                  warning-icon "~OK")))))

;;; Do the actual saving of the document data.
(defmethod DOC-SAVE-DATA ((doc document) file)
  (let ((doc-frame (doc-frame doc)))
    (fasl-write file (list (page-width doc-frame t)
                           (page-height doc-frame t)))
    (fasl-write file (window-exterior doc-frame))
    (fasl-write file (window-state doc-frame))
    (fasl-write file (multiple-value-list (scroll-range doc)))
    (fasl-write file (scroll-position doc))
    (fasl-write file (same-node-same-arg-on doc))
    (fasl-write file (dialog-create-on doc))
    (fasl-write file (grid-on doc))
    (fasl-write file (icon-list doc))))


;;;-------------------------------------------------------------------------;;;
;;;  D O C - S A V E  definition                                            ;;;
;;;-------------------------------------------------------------------------;;;

;;; Loading a document.
(defmethod DOC-LOAD ((filename pathname))
  (with-lock *prevent-actions*
    (with-open-file (file filename :direction :input)
      (if (and file (equal (fasl-read file) :booms-file))
        (let ((*loading-now* t)
              doc-type doc)
          (say "Loading...")
          (setf doc-type (fasl-read file))
          (setf doc (make-document doc-type
                                   :title        (pathname-to-string filename)
                                   :pathname     filename
                                   :window-state :shrunk))
          (doc-load-data doc file)
          (say "Done.")
          (setf (modified-p doc) nil)
          (close file))
        (my-pop-up-message-dialog *booms-app* "Booms Error"
                                  "Error loading document"
                                  warning-icon "~OK")))))

;;; Do the actual loading of the document data.
(defmethod DOC-LOAD-DATA ((doc document) file)
  (let* ((doc-frame             (doc-frame doc))
         (page-dimensions       (fasl-read file))
         (window-exterior       (fasl-read file))
         (window-state          (fasl-read file))
         (scroll-ranges         (fasl-read file))
         (scroll-position       (fasl-read file))
         (same-node-same-arg-on (fasl-read file))
         (dialog-create-on      (fasl-read file))
         (grid-on               (fasl-read file))
         (icon-list             (fasl-read file)))
    (setf (window-exterior doc-frame) window-exterior)
    (set-page-size doc-frame
                   (first page-dimensions) (second page-dimensions))
    (clear-page doc)
    (setf (window-exterior doc-frame) window-exterior)
    (set-scroll-range doc (first scroll-ranges) (second scroll-ranges))
    (scroll-to doc scroll-position)
    (setf (same-node-same-arg-on doc) same-node-same-arg-on)
    (setf (dialog-create-on      doc) dialog-create-on)
    (setf (grid-on               doc) grid-on)
    (setf (icon-list doc) icon-list)
    (dolist (icon icon-list)
      (setf (icon-doc icon) doc)
      (let ((*registering-type* t)) ; Prevent selecting this type.
        (new-named-abstraction icon)))
    (reset-view doc)
    (setf (window-state doc-frame) window-state)
    (select-doc doc)))


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :SAVELOAD)

